"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = transformRewriteImports;
exports.defaultRecognizedExtensions = void 0;
require("core-js/modules/es.array.push.js");
require("core-js/modules/esnext.iterator.constructor.js");
require("core-js/modules/esnext.iterator.for-each.js");
require("core-js/modules/esnext.iterator.some.js");
var _package = require("../../package.json");
var _debug = _interopRequireDefault(require("debug"));
var _template = _interopRequireDefault(require("@babel/template"));
var util = _interopRequireWildcard(require("@babel/types"));
function _getRequireWildcardCache(e) { if ("function" != typeof WeakMap) return null; var r = new WeakMap(), t = new WeakMap(); return (_getRequireWildcardCache = function (e) { return e ? t : r; })(e); }
function _interopRequireWildcard(e, r) { if (!r && e && e.__esModule) return e; if (null === e || "object" != typeof e && "function" != typeof e) return { default: e }; var t = _getRequireWildcardCache(r); if (t && t.has(e)) return t.get(e); var n = { __proto__: null }, a = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var u in e) if ("default" !== u && {}.hasOwnProperty.call(e, u)) { var i = a ? Object.getOwnPropertyDescriptor(e, u) : null; i && (i.get || i.set) ? Object.defineProperty(n, u, i) : n[u] = e[u]; } return n.default = e, t && t.set(e, n), n; }
function _interopRequireDefault(e) { return e && e.__esModule ? e : { default: e }; }
const debugNamespace = `${_package.name}:index`;
const debug = (0, _debug.default)(debugNamespace);
const globalMetadata = {};
const defaultRecognizedExtensions = exports.defaultRecognizedExtensions = ['.js', '.jsx', '.mjs', '.cjs', '.json'];
let reporterTimeout;
function transformRewriteImports() {
  let rewriteOptionsObjIdentifier = undefined;
  let rewriteFnIdentifier = undefined;
  return {
    name: 'transform-rewrite-imports',
    visitor: {
      Program: {
        enter() {
          rewriteOptionsObjIdentifier = undefined;
          rewriteFnIdentifier = undefined;
          clearTimeout(reporterTimeout);
        },
        exit(_, state) {
          if (process.env.NODE_ENV !== 'test') {
            reporterTimeout = setTimeout(() => {
              if (!state.opts.silent) {
                let totalGlobalImports = 0;
                let totalTransformedImports = 0;
                let totalFiles = 0;
                const globalMetadataEntries = Object.entries(globalMetadata);
                globalMetadataEntries.forEach(([path, metadata]) => {
                  if (state.opts.verbose) {
                    console.log(`└── Rewrote ${metadata.transformedImports.length} of ${metadata.totalImports} imports in file ${path}`);
                    metadata.transformedImports.forEach(transformedImport => {
                      console.log(`  » ${transformedImport}`);
                    });
                  }
                  totalGlobalImports += metadata.totalImports;
                  totalTransformedImports += metadata.transformedImports.length;
                  totalFiles += 1;
                });
                if (!state.opts.appendExtension && !state.opts.replaceExtensions) {
                  console.log(`(${_package.name} was loaded without any meaningful configuration, making it a noop)`);
                } else {
                  console.log(`${state.opts.verbose && globalMetadataEntries.length ? '  ---\n  ' : '└── '}Rewrote ${totalTransformedImports} of ${totalGlobalImports} imports ${totalFiles === 1 ? 'in 1 file' : `across ${totalFiles} files`}`);
                }
              }
            }, 100);
          }
        }
      },
      ImportDeclaration: declarationHandler,
      ExportAllDeclaration: declarationHandler,
      ExportNamedDeclaration: declarationHandler,
      TSImportType(path, state) {
        const {
          appendExtension,
          recognizedExtensions,
          replaceExtensions
        } = {
          ...state.opts,
          recognizedExtensions: state.opts.recognizedExtensions || defaultRecognizedExtensions
        };
        const {
          argument: specifier
        } = path.node;
        const specifierType = 'type-only dynamic import';
        const metadata = getLocalMetadata(state);
        metadata.totalImports += 1;
        if (appendExtension || replaceExtensions) {
          const filepath = getFilenameFromState(state);
          const importPath = specifier.value;
          const rewrittenPath = rewrite(importPath, {
            appendExtension,
            recognizedExtensions,
            replaceExtensions,
            filepath
          });
          if (importPath === rewrittenPath) {
            debug(`[%O]: %O evaluated but unchanged: "%O"`, filepath, specifierType, importPath);
          } else {
            path.node.argument = util.stringLiteral(rewrittenPath);
            const debugString = `[TSImportType]: ${specifierType} "${importPath}" => "${rewrittenPath}"`;
            metadata.transformedImports.push(debugString);
            debug(debugString);
          }
        }
      },
      CallExpression(path, state) {
        const isDynamicImport = path.node.callee?.type === 'Import';
        const isRequire = path.node.callee?.type === 'Identifier' && path.node.callee?.name === 'require';
        const firstArgument = path.node.arguments?.[0];
        const firstArgumentIsStringLiteral = firstArgument?.type === 'StringLiteral';
        const {
          appendExtension,
          recognizedExtensions,
          replaceExtensions
        } = {
          ...state.opts,
          recognizedExtensions: state.opts.recognizedExtensions || defaultRecognizedExtensions
        };
        if (isDynamicImport || isRequire) {
          const metadata = getLocalMetadata(state);
          metadata.totalImports += 1;
          if (appendExtension || replaceExtensions) {
            const specifierType = isDynamicImport ? 'dynamic import' : 'require';
            const filepath = getFilenameFromState(state);
            if (!firstArgument) {
              console.warn(`[${filepath}]: warning: a ${specifierType} statement has no arguments!`);
            } else if (firstArgumentIsStringLiteral) {
              const importPath = firstArgument.value;
              const rewrittenPath = rewrite(importPath, {
                appendExtension,
                recognizedExtensions,
                replaceExtensions,
                filepath
              });
              if (importPath === rewrittenPath) {
                debug(`[%O]: %O evaluated but unchanged: "%O"`, filepath, specifierType, importPath);
              } else {
                path.node.arguments[0] = util.stringLiteral(rewrittenPath);
                const debugString = `[CallExpression]: ${specifierType} "${importPath}" => "${rewrittenPath}"`;
                metadata.transformedImports.push(debugString);
                debug(debugString);
              }
            } else {
              const globalScope = path.scope.getProgramParent();
              if (!rewriteFnIdentifier) {
                const injectedRewriter = rewrite.toString().replaceAll(/\boptions\.filepath\b/g, `'${filepath}'`);
                debug('[CallExpression]: injectedRewriter: %O', injectedRewriter);
                const rewriteFnAst = _template.default.expression.ast(injectedRewriter);
                rewriteFnIdentifier = globalScope.generateUidIdentifier('_rewrite');
                globalScope.push({
                  id: rewriteFnIdentifier,
                  init: rewriteFnAst,
                  kind: 'const'
                });
              }
              if (!rewriteOptionsObjIdentifier) {
                const functionExpressionCache = {};
                let rewriteOptionsObjString = JSON.stringify({
                  appendExtension,
                  recognizedExtensions,
                  replaceExtensions
                }, (_key, value) => {
                  if (typeof value === 'function') {
                    const fnExprString = value.toString();
                    const fnExprId = `%&^#%%${_package.name}%%${Object.keys(functionExpressionCache).length + 1}#^&%`;
                    functionExpressionCache[fnExprId] = fnExprString;
                    return fnExprId;
                  }
                  return value;
                });
                Object.entries(functionExpressionCache).forEach(([fnExprId, fnExprString]) => {
                  rewriteOptionsObjString = rewriteOptionsObjString.replaceAll(`"${fnExprId}"`, fnExprString);
                });
                const rewriteOptionsObjAst = _template.default.expression.ast(rewriteOptionsObjString);
                rewriteOptionsObjIdentifier = globalScope.generateUidIdentifier('_rewrite_options');
                globalScope.push({
                  id: rewriteOptionsObjIdentifier,
                  init: rewriteOptionsObjAst,
                  kind: 'const'
                });
              }
              path.get('arguments')[0].replaceWith(util.callExpression(rewriteFnIdentifier, [firstArgument, rewriteOptionsObjIdentifier]));
              const debugString = `[CallExpression]: ${specifierType} first argument wrapped with rewrite function`;
              metadata.transformedImports.push(debugString);
              debug(debugString);
            }
          }
        }
      }
    }
  };
}
function declarationHandler(path, state) {
  const metadata = getLocalMetadata(state);
  metadata.totalImports += 1;
  if (path.node.source && (state.opts.appendExtension || state.opts.replaceExtensions)) {
    const importPath = path.node.source.value;
    const specifierType = path.node.type.startsWith('Import') ? 'import' : 'export';
    const filepath = getFilenameFromState(state);
    const rewrittenPath = rewrite(importPath, {
      ...state.opts,
      recognizedExtensions: state.opts.recognizedExtensions || defaultRecognizedExtensions,
      filepath
    });
    if (importPath === rewrittenPath) {
      debug(`[${getFilenameFromState(state)}]: ${specifierType} evaluated but unchanged: "${importPath}"`);
    } else {
      path.node.source = util.stringLiteral(rewrittenPath);
      const debugString = `[declarationHandler]: ${specifierType} "${importPath}" => "${rewrittenPath}"`;
      metadata.transformedImports.push(debugString);
      debug(debugString);
    }
  }
}
function getLocalMetadata(state) {
  const key = getFilenameFromState(state);
  return globalMetadata[key] = globalMetadata[key] || {
    totalImports: 0,
    transformedImports: []
  };
}
function getFilenameFromState(state) {
  return state.filename || '<no filepath was set in your babel configuration>';
}
const rewrite = (specifier, options) => {
  if (typeof specifier !== 'string') {
    throw new TypeError(`rewrite error: expected specifier of type string, not ${typeof specifier}`);
  }
  let replacementMap;
  replacementMap = undefined;
  if (options.replaceExtensions) {
    Object.entries(options.replaceExtensions).some(([target, replacement]) => {
      if (target.startsWith('^') || target.endsWith('$')) {
        const targetRegExp = new RegExp(target);
        const capturingGroups = Array.from(specifier.match(targetRegExp) || []);
        if (capturingGroups.length) {
          replacementMap = [targetRegExp, replacement, capturingGroups];
          return true;
        }
      } else if (specifier.endsWith(target)) {
        replacementMap = [target, replacement, []];
        return true;
      }
    });
  }
  let finalImportPath = specifier;
  if (replacementMap) {
    const [target, replacement, capturingGroups] = replacementMap;
    finalImportPath = finalImportPath.replace(target, typeof replacement === 'string' ? replacement : replacement({
      specifier,
      capturingGroups,
      filepath: options.filepath
    }));
  }
  const isRelative = finalImportPath.startsWith('./') || finalImportPath.startsWith('.\\') || finalImportPath.startsWith('../') || finalImportPath.startsWith('..\\') || finalImportPath === '.' || finalImportPath === '..';
  if (options.appendExtension && isRelative) {
    const endsWithSlash = /(\/|\\)$/.test(finalImportPath);
    const basenameIsDots = /(^\.?\.(\/|\\)?$)|((\/|\\)\.?\.(\/|\\)?$)/.test(finalImportPath);
    const extensionToAppend = typeof options.appendExtension === 'string' ? options.appendExtension : options.appendExtension({
      specifier,
      capturingGroups: [],
      filepath: options.filepath
    });
    if (extensionToAppend !== undefined) {
      if (basenameIsDots) {
        finalImportPath += `${endsWithSlash ? '' : '/'}index${extensionToAppend}`;
      } else {
        const hasRecognizedExtension = !endsWithSlash && options.recognizedExtensions.some(extension => {
          return finalImportPath.endsWith(extension);
        });
        if (!hasRecognizedExtension) {
          finalImportPath = endsWithSlash ? finalImportPath + `index${extensionToAppend}` : finalImportPath + extensionToAppend;
        }
      }
    }
  }
  return finalImportPath;
};